/*
 *
 *
 *    cfsk.h
 *
 *    FSK demodulator.
 *
 *    Copyright (C) 2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __CFSK_H
#define __CFSK_H

#include <stdint.h>
#include "cavg.h"
#include "ctrthr.h"
#include "sign.h"

#define FIXED_POINT_FACTOR (256)
#define FIXED_POINT_SHIFT (8)

//
//  Reference:
//     FSK Modulation and Demodulation With the MSP430 Microcontroller (TI)
//     http://www.ti.com/lit/an/slaa037/slaa037.pdf
//

namespace KK5JY {
	namespace DSP {

		//
		//  Demodulator class
		//
		class DelayLineFskDemod {
			public:
				struct Config {
					// center frequency
					double f0;
					// sampling frequency
					uint16_t fs;
					// integrator length
					uint16_t int_len;
					// counter rail
					uint8_t ctr_limit;
					// counter trigger level
					uint8_t ctr_trigger;

					Config() {
						f0 = 1000;
						fs = 8000;
						int_len = 100;
						ctr_limit = 10;
						ctr_trigger = 6;
					}
				};
			public:
				// the exact value of the quarter-wavelength delay line
				const double N;
				// the length of the delay array
				const uint16_t length;
			private:
				// the whole part of N
				const uint16_t idx;
				// the fraction * FACTOR
				const uint16_t frac;

				// the delay line
				int8_t * const delay;
				// a pointer for walking the delay line
				int8_t * ptr;
				// the integrator - int16 samples, int32 accumulator
				Averager<int16_t, int32_t> integrator;
				// the output hysteresis counter
				CounterThreshold<int8_t, int8_t> ctr;

			public:
				DelayLineFskDemod(const Config &c);

			public:
				// run one sample through the demodulator
				bool run(int8_t sample);

				// return the integrator length
				size_t avgSize() const { return integrator.size(); }

				// return the current integrated deviation (mostly for alignment)
				int16_t deviation() const { return integrator.value(); }

				// return the current output counter value
				int8_t counter() const { return ctr.value(); }
		};

		//
		//  ctor
		//
		inline DelayLineFskDemod::DelayLineFskDemod(const Config &c)
			: N(c.fs / (4.0 * c.f0)),						// the exact value of N
			  length(static_cast<uint16_t>(round(N + 1))),	// history length
			  idx(static_cast<uint16_t>(N)),				// the whole part of N
			  frac(static_cast<double>(FIXED_POINT_FACTOR)
			        * (N - static_cast<uint16_t>(N))),		// the fractional part of N, * FACTOR
			  delay(new int8_t[length]),					// the delay array
			  integrator(c.int_len),						// the integrator
			  ctr(c.ctr_limit, c.ctr_trigger)				// output hysteresis counter
		{
			// initialize the quadrature delay line
			for (uint16_t i = 0; i != length; ++i) {
				delay[i] = 0;
			}
			ptr = delay;
		}

		//
		//  run
		//
		inline bool DelayLineFskDemod::run(int8_t sample) {
			// fetch the quadrature sample
			int8_t *q = ptr + idx;
			if (q >= (delay + length)) {
				q -= length;
			}
			int8_t q1 = *q;

			// and interpolate, if needed
			if (frac) {
				q = ptr + static_cast<uint8_t>(N) + 1;
				if (q >= (delay + length)) {
					q -= length;
				}
				int8_t q2 = *q;
				q1 = (
						(q2 * frac) + (q1 * (FIXED_POINT_FACTOR - static_cast<int16_t>(frac)))
					 ) >> FIXED_POINT_SHIFT;
			}

			// compute the product of the the signal and its quadrature;
			//   then run the integrator on the product
			integrator.add(sample * q1);

			// move the current pointer backwards through the delay line
			*ptr-- = sample;
			if (ptr == (delay - 1)) {
				ptr = delay + length - 1;
			}

			// run output hysteresis
			return ctr.run(sign(integrator.accum()));
		}
	}
}

#endif
